@@ -61,9 +61,10 @@ _pdfioFileAddMappedObj(
   pdf->num_objmaps ++;
 
   map->obj        = dst_obj;
-  map->src_pdf    = src_obj->pdf;
   map->src_number = src_obj->number;
 
+  memcpy(map->src_id, src_obj->pdf->file_id, sizeof(map->src_id));
+
   // Sort as needed...
   if (pdf->num_objmaps > 1 && compare_objmaps(map, pdf->objmaps + pdf->num_objmaps - 2) < 0)
     qsort(pdf->objmaps, pdf->num_objmaps, sizeof(_pdfio_objmap_t), (int (*)(const void *, const void *))compare_objmaps);
@@ -181,8 +182,9 @@ pdfioFileClose(pdfio_file_t *pdf)	// I - PDF file
 // of 8.27x11in (the intersection of US Letter and ISO A4) is used.
 //
 // The "error_cb" and "error_cbdata" arguments specify an error handler callback
-// and its data pointer - if `NULL` the default error handler is used that
-// writes error messages to `stderr`.
+// and its data pointer - if `NULL` then the default error handler is used that
+// writes error messages to `stderr`.  The error handler callback should return
+// `true` to continue writing the PDF file or `false` to stop.
 //
 
 pdfio_file_t *				// O - PDF file or `NULL` on error
@@ -260,6 +262,8 @@ pdfioFileCreateArrayObj(
 // This function creates a new object with a name value in a PDF file.
 // You must call @link pdfioObjClose@ to write the object to the file.
 //
+// @since PDFio v1.4@
+//
 
 pdfio_obj_t *				// O - New object
 pdfioFileCreateNameObj(
@@ -289,6 +293,8 @@ pdfioFileCreateNameObj(
 // This function creates a new object with a number value in a PDF file.
 // You must call @link pdfioObjClose@ to write the object to the file.
 //
+// @since PDFio v1.2@
+//
 
 pdfio_obj_t *				// O - New object
 pdfioFileCreateNumberObj(
@@ -415,8 +421,9 @@ _pdfioFileCreateObj(
 // of 8.27x11in (the intersection of US Letter and ISO A4) is used.
 //
 // The "error_cb" and "error_cbdata" arguments specify an error handler callback
-// and its data pointer - if `NULL` the default error handler is used that
-// writes error messages to `stderr`.
+// and its data pointer - if `NULL` then the default error handler is used that
+// writes error messages to `stderr`.  The error handler callback should return
+// `true` to continue writing the PDF file or `false` to stop.
 //
 // > *Note*: Files created using this API are slightly larger than those
 // > created using the @link pdfioFileCreate@ function since stream lengths are
@@ -518,6 +525,8 @@ pdfioFileCreatePage(pdfio_file_t *pdf,	// I - PDF file
 // This function creates a new object with a string value in a PDF file.
 // You must call @link pdfioObjClose@ to write the object to the file.
 //
+// @since PDFio v1.2@
+//
 
 pdfio_obj_t *				// O - New object
 pdfioFileCreateStringObj(
@@ -669,7 +678,7 @@ _pdfioFileFindMappedObj(
     return (NULL);
 
   // Otherwise search for a match...
-  key.src_pdf    = src_pdf;
+  memcpy(key.src_id, src_pdf->file_id, sizeof(key.src_id));
   key.src_number = src_number;
 
   if ((match = (_pdfio_objmap_t *)bsearch(&key, pdf->objmaps, pdf->num_objmaps, sizeof(_pdfio_objmap_t), (int (*)(const void *, const void *))compare_objmaps)) != NULL)
@@ -987,8 +996,14 @@ pdfioFileGetVersion(
 // PDF file requires a password, the open will always fail.
 //
 // The "error_cb" and "error_cbdata" arguments specify an error handler callback
-// and its data pointer - if `NULL` the default error handler is used that
-// writes error messages to `stderr`.
+// and its data pointer - if `NULL` then the default error handler is used that
+// writes error messages to `stderr`.  The error handler callback should return
+// `true` to continue reading the PDF file or `false` to stop.
+//
+// > Note: Error messages starting with "WARNING:" are actually warning
+// > messages - the callback should normally return `true` to allow PDFio to
+// > try to resolve the issue.  In addition, some errors are unrecoverable and
+// > ignore the return value of the error callback.
 //
 
 pdfio_file_t *				// O - PDF file
@@ -1006,6 +1021,10 @@ pdfioFileOpen(
 		*end;			// End of line
   ssize_t	bytes;			// Bytes read
   off_t		xref_offset;		// Offset to xref table
+  time_t	curtime;		// Creation date/time
+  unsigned char	*id_value;		// ID value
+  size_t	id_valuelen;		// Length of ID value
+  _pdfio_sha256_t ctx;			// Hashing context
 
 
   PDFIO_DEBUG("pdfioFileOpen(filename=\"%s\", password_cb=%p, password_cbdata=%p, error_cb=%p, error_cbdata=%p)\n", filename, (void *)password_cb, (void *)password_cbdata, (void *)error_cb, (void *)error_cbdata);
@@ -1049,7 +1068,7 @@ pdfioFileOpen(
   }
 
   // Read the header from the first line...
-  if (!_pdfioFileGets(pdf, line, sizeof(line)))
+  if (!_pdfioFileGets(pdf, line, sizeof(line), true))
     goto error;
 
   if ((strncmp(line, "%PDF-1.", 7) && strncmp(line, "%PDF-2.", 7)) || !isdigit(line[7] & 255))
@@ -1063,7 +1082,7 @@ pdfioFileOpen(
   pdf->version = strdup(line + 5);
 
   // Grab the last 1k of the file to find the start of the xref table...
-  if (_pdfioFileSeek(pdf, -1024, SEEK_END) < 0)
+  if (_pdfioFileSeek(pdf, 1 - sizeof(line), SEEK_END) < 0)
   {
     _pdfioFileError(pdf, "Unable to read startxref data.");
     goto error;
@@ -1075,28 +1094,48 @@ pdfioFileOpen(
     goto error;
   }
 
+  PDFIO_DEBUG("pdfioOpen: Read %d bytes at end of file.\n", (int)bytes);
+
   line[bytes] = '\0';
   end = line + bytes - 9;
 
   for (ptr = line; ptr < end; ptr ++)
   {
-    if (!memcmp(ptr, "startxref", 9))
+    if (!strncmp(ptr, "startxref", 9) && !strstr(ptr + 9, "startxref") && strtol(ptr + 9, NULL, 10) > 0)
       break;
   }
 
   if (ptr >= end)
   {
-    _pdfioFileError(pdf, "Unable to find start of xref table.");
-    goto error;
-  }
+    if (!_pdfioFileError(pdf, "WARNING: Unable to find start of cross-reference table, will attempt to rebuild."))
+      goto error;
 
-  xref_offset = (off_t)strtol(ptr + 9, NULL, 10);
-
-  if (!load_xref(pdf, xref_offset, password_cb, password_cbdata))
-  {
     if (!repair_xref(pdf, password_cb, password_cbdata))
       goto error;
   }
+  else
+  {
+    PDFIO_DEBUG("pdfioFileOpen: line=%p,ptr=%p(\"%s\")\n", line, ptr, ptr);
+
+    xref_offset = (off_t)strtol(ptr + 9, NULL, 10);
+
+    PDFIO_DEBUG("pdfioFileOpen: xref_offset=%lu\n", (unsigned long)xref_offset);
+
+    if (!load_xref(pdf, xref_offset, password_cb, password_cbdata))
+      goto error;
+  }
+
+  // Create the unique file identifier string for the object map...
+  curtime = pdfioFileGetCreationDate(pdf);
+
+  _pdfioCryptoSHA256Init(&ctx);
+  _pdfioCryptoSHA256Append(&ctx, (uint8_t *)pdf->filename, strlen(pdf->filename));
+  _pdfioCryptoSHA256Append(&ctx, (uint8_t *)&curtime, sizeof(curtime));
+  if ((id_value = pdfioArrayGetBinary(pdf->id_array, 0, &id_valuelen)) != NULL)
+    _pdfioCryptoSHA256Append(&ctx, id_value, id_valuelen);
+  if ((id_value = pdfioArrayGetBinary(pdf->id_array, 1, &id_valuelen)) != NULL)
+    _pdfioCryptoSHA256Append(&ctx, id_value, id_valuelen);
+  _pdfioCryptoSHA256Finish(&ctx, pdf->file_id);
 
   return (pdf);
 
@@ -1210,6 +1249,8 @@ pdfioFileSetPermissions(
   if (encryption == PDFIO_ENCRYPTION_NONE)
     return (true);
 
+  pdf->encrypt_metadata = true;
+
   return (_pdfioCryptoLock(pdf, permissions, encryption, owner_password, user_password));
 }
 
@@ -1350,10 +1391,11 @@ static int				// O - Result of comparison
 compare_objmaps(_pdfio_objmap_t *a,	// I - First object map
                 _pdfio_objmap_t *b)	// I - Second object map
 {
-  if (a->src_pdf < b->src_pdf)
-    return (-1);
-  else if (a->src_pdf > b->src_pdf)
-    return (1);
+  int	ret = memcmp(a->src_id, b->src_id, sizeof(a->src_id));
+					// Result of comparison
+
+  if (ret)
+    return (ret);
   else if (a->src_number < b->src_number)
     return (-1);
   else if (a->src_number > b->src_number)
@@ -1382,6 +1424,8 @@ create_common(
   pdfio_file_t	*pdf;			// New PDF file
   pdfio_dict_t	*dict;			// Dictionary
   unsigned char	id_value[16];		// File ID value
+  time_t	curtime;		// Creation date/time
+  _pdfio_sha256_t ctx;			// Hashing context
 
 
   PDFIO_DEBUG("create_common(filename=\"%s\", fd=%d, output_cb=%p, output_cbdata=%p, version=\"%s\", media_box=%p, crop_box=%p, error_cb=%p, error_cbdata=%p)\n", filename, fd, (void *)output_cb, (void *)output_cbdata, version, (void *)media_box, (void *)crop_box, (void *)error_cb, (void *)error_cbdata);
@@ -1472,7 +1516,9 @@ create_common(
   if ((dict = pdfioDictCreate(pdf)) == NULL)
     goto error;
 
-  pdfioDictSetDate(dict, "CreationDate", time(NULL));
+  curtime = time(NULL);
+
+  pdfioDictSetDate(dict, "CreationDate", curtime);
   pdfioDictSetString(dict, "Producer", "pdfio/" PDFIO_VERSION);
 
   if ((pdf->info_obj = pdfioFileCreateObj(pdf, dict)) == NULL)
@@ -1497,6 +1543,14 @@ create_common(
     pdfioArrayAppendBinary(pdf->id_array, id_value, sizeof(id_value));
   }
 
+  // Create the unique file identifier string for the object map...
+  _pdfioCryptoSHA256Init(&ctx);
+  _pdfioCryptoSHA256Append(&ctx, (uint8_t *)pdf->filename, strlen(pdf->filename));
+  _pdfioCryptoSHA256Append(&ctx, (uint8_t *)&curtime, sizeof(curtime));
+  _pdfioCryptoSHA256Append(&ctx, id_value, sizeof(id_value));
+  _pdfioCryptoSHA256Append(&ctx, id_value, sizeof(id_value));
+  _pdfioCryptoSHA256Finish(&ctx, pdf->file_id);
+
   return (pdf);
 
   // Common error handling code...
@@ -1675,7 +1729,10 @@ load_pages(pdfio_file_t *pdf,		// I - PDF file
   }
 
   if ((type = pdfioDictGetName(dict, "Type")) == NULL || (strcmp(type, "Pages") && strcmp(type, "Page")))
-    return (false);
+  {
+    if (!_pdfioFileError(pdf, "WARNING: No Type value for pages object."))
+      return (false);
+  }
 
   // If there is a Kids array, then this is a parent node and we have to look
   // at the child objects...
@@ -1742,31 +1799,32 @@ load_xref(
   int		generation;		// Generation number
   _pdfio_token_t tb;			// Token buffer/stack
   off_t		line_offset;		// Offset to start of line
+  pdfio_obj_t	*pages_obj;		// Pages object
 
 
   while (!done)
   {
     if (_pdfioFileSeek(pdf, xref_offset, SEEK_SET) != xref_offset)
     {
-      _pdfioFileError(pdf, "Unable to seek to start of xref table.");
-      return (false);
+      PDFIO_DEBUG("load_xref: Unable to seek to %lu.\n", (unsigned long)xref_offset);
+      goto repair;
     }
 
     do
     {
       line_offset = _pdfioFileTell(pdf);
 
-      if (!_pdfioFileGets(pdf, line, sizeof(line)))
+      if (!_pdfioFileGets(pdf, line, sizeof(line), true))
       {
-	_pdfioFileError(pdf, "Unable to read start of xref table.");
-	return (false);
+	PDFIO_DEBUG("load_xref: Unable to read line at offset %lu.\n", (unsigned long)line_offset);
+        goto repair;
       }
     }
     while (!line[0]);
 
     PDFIO_DEBUG("load_xref: line_offset=%lu, line='%s'\n", (unsigned long)line_offset, line);
 
-    if (isdigit(line[0] & 255) && strlen(line) > 4 && (!strcmp(line + strlen(line) - 4, " obj") || ((ptr = strstr(line, " obj")) != NULL && ptr[4] == '<')))
+    if (isdigit(line[0] & 255) && strlen(line) > 4 && (!strcmp(line + strlen(line) - 4, " obj") || ((ptr = strstr(line, " obj")) != NULL && (ptr[4] == '<' || isspace(ptr[4])))))
     {
       // Cross-reference stream
       pdfio_obj_t	*obj;		// Object
@@ -1788,14 +1846,14 @@ load_xref(
 
       if ((number = strtoimax(line, &ptr, 10)) < 1)
       {
-	_pdfioFileError(pdf, "Bad xref table header '%s'.", line);
-	return (false);
+        PDFIO_DEBUG("load_xref: Unable to scan object number.\n");
+        goto repair;
       }
 
       if ((generation = (int)strtol(ptr, &ptr, 10)) < 0 || (generation > 65535 && number != 0))
       {
-	_pdfioFileError(pdf, "Bad xref table header '%s'.", line);
-	return (false);
+        PDFIO_DEBUG("load_xref: Unable to scan generation number (%u).\n", (unsigned)generation);
+        goto repair;
       }
 
       while (isspace(*ptr & 255))
@@ -1803,14 +1861,14 @@ load_xref(
 
       if (strncmp(ptr, "obj", 3))
       {
-	_pdfioFileError(pdf, "Bad xref table header '%s'.", line);
-	return (false);
+        PDFIO_DEBUG("load_xref: No 'obj' after object number and generation (saw '%s').\n", ptr);
+        goto repair;
       }
 
       if (_pdfioFileSeek(pdf, line_offset + (off_t)(ptr + 3 - line), SEEK_SET) < 0)
       {
-        _pdfioFileError(pdf, "Unable to seek to xref object %lu %u.", (unsigned long)number, (unsigned)generation);
-        return (false);
+        PDFIO_DEBUG("load_xref: Unable to seek to start of cross-reference object dictionary.\n");
+        goto repair;
       }
 
       PDFIO_DEBUG("load_xref: Loading object %lu %u.\n", (unsigned long)number, (unsigned)generation);
@@ -1825,21 +1883,21 @@ load_xref(
 
       if (!_pdfioValueRead(pdf, obj, &tb, &trailer, 0))
       {
-        _pdfioFileError(pdf, "Unable to read cross-reference stream dictionary.");
-        return (false);
+        PDFIO_DEBUG("load_xref: Unable to read cross-reference object dictionary.\n");
+        goto repair;
       }
       else if (trailer.type != PDFIO_VALTYPE_DICT)
       {
-	_pdfioFileError(pdf, "Cross-reference stream does not have a dictionary.");
-	return (false);
+        PDFIO_DEBUG("load_xref: Expected dictionary for cross-reference object (type=%d).", trailer.type);
+        goto repair;
       }
 
       obj->value = trailer;
 
       if (!_pdfioTokenGet(&tb, line, sizeof(line)) || strcmp(line, "stream"))
       {
-        _pdfioFileError(pdf, "Unable to get stream after xref dictionary.");
-        return (false);
+        PDFIO_DEBUG("load_xref: No stream token after dictionary (got '%s').\n", line);
+        goto repair;
       }
 
       PDFIO_DEBUG("load_xref: tb.bufptr=%p, tb.bufend=%p, tb.bufptr[0]=0x%02x, tb.bufptr[0]=0x%02x\n", tb.bufptr, tb.bufend, tb.bufptr[0], tb.bufptr[1]);
@@ -1857,8 +1915,8 @@ load_xref(
 
       if ((w_array = pdfioDictGetArray(trailer.value.dict, "W")) == NULL)
       {
-	_pdfioFileError(pdf, "Cross-reference stream does not have required W key.");
-	return (false);
+        PDFIO_DEBUG("load_xref: Missing W array in cross-reference objection dictionary.\n");
+        goto repair;
       }
 
       w[0]    = (size_t)pdfioArrayGetNumber(w_array, 0);
@@ -1868,16 +1926,18 @@ load_xref(
       w_2     = w[0];
       w_3     = w[0] + w[1];
 
-      if (w[1] == 0 || w[2] > 4 || w[0] > sizeof(buffer) || w[1] > sizeof(buffer) || w[2] > sizeof(buffer) || w_total > sizeof(buffer))
+      PDFIO_DEBUG("W=[%u %u %u], w_total=%u\n", (unsigned)w[0], (unsigned)w[1], (unsigned)w[2], (unsigned)w_total);
+
+      if (pdfioArrayGetSize(w_array) > 3 || w[1] == 0 || w[2] > 4 || w[0] > sizeof(buffer) || w[1] > sizeof(buffer) || w[2] > sizeof(buffer) || w_total > sizeof(buffer))
       {
-	_pdfioFileError(pdf, "Cross-reference stream has invalid W key [%u %u %u].", (unsigned)w[0], (unsigned)w[1], (unsigned)w[2]);
-	return (false);
+        PDFIO_DEBUG("load_xref: Bad W array in cross-reference objection dictionary.\n");
+        goto repair;
       }
 
       if ((st = pdfioObjOpenStream(obj, true)) == NULL)
       {
-	_pdfioFileError(pdf, "Unable to open cross-reference stream.");
-	return (false);
+        PDFIO_DEBUG("load_xref: Unable to open cross-reference stream.\n");
+        goto repair;
       }
 
       for (index_n = 0; index_n < index_count; index_n += 2)
@@ -1897,7 +1957,20 @@ load_xref(
 	{
 	  count --;
 
-	  PDFIO_DEBUG("load_xref: number=%u %02X%02X%02X%02X%02X\n", (unsigned)number, buffer[0], buffer[1], buffer[2], buffer[3], buffer[4]);
+#ifdef DEBUG
+          if (w_total > 5)
+	    PDFIO_DEBUG("load_xref: number=%u %02X%02X%02X%02X%02X...\n", (unsigned)number, buffer[0], buffer[1], buffer[2], buffer[3], buffer[4]);
+	  else if (w_total == 5)
+	    PDFIO_DEBUG("load_xref: number=%u %02X%02X%02X%02X%02X\n", (unsigned)number, buffer[0], buffer[1], buffer[2], buffer[3], buffer[4]);
+	  else if (w_total == 4)
+	    PDFIO_DEBUG("load_xref: number=%u %02X%02X%02X%02X\n", (unsigned)number, buffer[0], buffer[1], buffer[2], buffer[3]);
+	  else if (w_total == 3)
+	    PDFIO_DEBUG("load_xref: number=%u %02X%02X%02X\n", (unsigned)number, buffer[0], buffer[1], buffer[2]);
+	  else if (w_total == 2)
+	    PDFIO_DEBUG("load_xref: number=%u %02X%02X\n", (unsigned)number, buffer[0], buffer[1]);
+	  else
+	    PDFIO_DEBUG("load_xref: number=%u %02X\n", (unsigned)number, buffer[0]);
+#endif // DEBUG
 
 	  // Check whether this is an object definition...
 	  if (w[0] > 0)
@@ -1979,6 +2052,7 @@ load_xref(
 	      else
 	      {
 		_pdfioFileError(pdf, "Too many object streams.");
+		pdfioStreamClose(st);
 	        return (false);
 	      }
 	    }
@@ -1987,7 +2061,10 @@ load_xref(
 	  {
 	    // Add this object...
 	    if (!add_obj(pdf, (size_t)number, (unsigned short)generation, (off_t)offset))
+	    {
+	      pdfioStreamClose(st);
 	      return (false);
+	    }
 	  }
 
 	  number ++;
@@ -2035,7 +2112,7 @@ load_xref(
 					// Offset of current line
 
       PDFIO_DEBUG("load_xref: Reading xref table starting at offset %lu\n", (unsigned long)trailer_offset);
-      while (_pdfioFileGets(pdf, line, sizeof(line)))
+      while (_pdfioFileGets(pdf, line, sizeof(line), false))
       {
         PDFIO_DEBUG("load_xref: '%s' at offset %lu\n", line, (unsigned long)trailer_offset);
 
@@ -2060,8 +2137,8 @@ load_xref(
 
 	if (sscanf(line, "%jd%jd", &number, &num_objects) != 2)
 	{
-	  _pdfioFileError(pdf, "Malformed xref table section '%s'.", line);
-	  return (false);
+	  PDFIO_DEBUG("load_xref: Unable to scan START COUNT from line.\n");
+	  goto repair;
 	}
 
 	// Read this group of objects...
@@ -2069,41 +2146,45 @@ load_xref(
 	{
 	  // Read a line from the file and validate it...
 	  if (_pdfioFileRead(pdf, line, 20) != 20)
-	    return (false);
+	  {
+	    PDFIO_DEBUG("load_xref: Unable to read 20 byte xref record.\n");
+	    goto repair;
+	  }
 
 	  line[20] = '\0';
 
-	  if (strcmp(line + 18, "\r\n") && strcmp(line + 18, " \n") && strcmp(line + 18, " \r"))
+	  if (strcmp(line + 18, "\r\n") && strcmp(line + 18, "\r\r") && strcmp(line + 18, " \n") && strcmp(line + 18, " \r"))
 	  {
-	    _pdfioFileError(pdf, "Malformed xref table entry '%s'.", line);
-	    return (false);
+	    PDFIO_DEBUG("load_xref: Bad end-of-line <%02X%02X>\n", line[18], line[19]);
+	    goto repair;
 	  }
+
 	  line[18] = '\0';
 
 	  // Parse the line
 	  if ((offset = strtoimax(line, &ptr, 10)) < 0)
 	  {
-	    _pdfioFileError(pdf, "Malformed xref table entry '%s'.", line);
-	    return (false);
+	    PDFIO_DEBUG("load_xref: Unable to scan offset.\n");
+	    goto repair;
 	  }
 
 	  if ((generation = (int)strtol(ptr, &ptr, 10)) < 0 || (generation > 65535 && offset != 0))
 	  {
-	    _pdfioFileError(pdf, "Malformed xref table entry '%s'.", line);
-	    return (false);
+	    PDFIO_DEBUG("load_xref: Unable to scan generation (%u).\n", (unsigned)generation);
+	    goto repair;
 	  }
 
 	  if (*ptr != ' ')
 	  {
-	    _pdfioFileError(pdf, "Malformed xref table entry '%s'.", line);
-	    return (false);
+	    PDFIO_DEBUG("load_xref: Missing space before type.\n");
+	    goto repair;
 	  }
 
 	  ptr ++;
 	  if (*ptr != 'f' && *ptr != 'n')
 	  {
-	    _pdfioFileError(pdf, "Malformed xref table entry '%s'.", line);
-	    return (false);
+	    PDFIO_DEBUG("load_xref: Bad type '%c'.\n", *ptr);
+	    goto repair;
 	  }
 
 	  if (*ptr == 'f')
@@ -2122,21 +2203,21 @@ load_xref(
 
       if (strncmp(line, "trailer", 7))
       {
-	_pdfioFileError(pdf, "Missing trailer.");
-	return (false);
+        PDFIO_DEBUG("load_xref: No trailer after xref table.\n");
+	goto repair;
       }
 
       _pdfioTokenInit(&tb, pdf, (_pdfio_tconsume_cb_t)_pdfioFileConsume, (_pdfio_tpeek_cb_t)_pdfioFilePeek, pdf);
 
       if (!_pdfioValueRead(pdf, NULL, &tb, &trailer, 0))
       {
-	_pdfioFileError(pdf, "Unable to read trailer dictionary.");
-	return (false);
+        PDFIO_DEBUG("load_xref: Unable to read trailer dictionary.\n");
+	goto repair;
       }
       else if (trailer.type != PDFIO_VALTYPE_DICT)
       {
-	_pdfioFileError(pdf, "Trailer is not a dictionary.");
-	return (false);
+        PDFIO_DEBUG("load_xref: Trailer not a dictionary (type=%d).\n", trailer.type);
+	goto repair;
       }
 
       PDFIO_DEBUG("load_xref: Got trailer dict.\n");
@@ -2158,8 +2239,7 @@ load_xref(
     }
     else
     {
-      _pdfioFileError(pdf, "Bad xref table header '%s'.", line);
-      return (false);
+      goto repair;
     }
 
     PDFIO_DEBUG("load_xref: Contents of trailer dictionary:\n");
@@ -2188,13 +2268,31 @@ load_xref(
 
   if ((pdf->root_obj = pdfioDictGetObj(pdf->trailer_dict, "Root")) == NULL)
   {
-    _pdfioFileError(pdf, "Missing Root object.");
-    return (false);
+    PDFIO_DEBUG("load_xref: Missing Root object.\n");
+    goto repair;
   }
 
   PDFIO_DEBUG("load_xref: Root=%p(%lu)\n", pdf->root_obj, (unsigned long)pdf->root_obj->number);
 
-  return (load_pages(pdf, pdfioDictGetObj(pdfioObjGetDict(pdf->root_obj), "Pages"), 0));
+  if ((pages_obj = pdfioDictGetObj(pdfioObjGetDict(pdf->root_obj), "Pages")) == NULL)
+  {
+    PDFIO_DEBUG("load_xref: Missing Pages object.\n");
+    goto repair;
+  }
+
+  PDFIO_DEBUG("load_xref: Pages=%p(%lu)\n", pdf->root_obj, (unsigned long)pdf->root_obj->number);
+
+  return (load_pages(pdf, pages_obj, 0));
+
+  // If we get here the cross-reference table is busted - try repairing if the
+  // error callback says to proceed...
+
+  repair:
+
+  if (_pdfioFileError(pdf, "WARNING: Cross-reference is damaged, will attempt to rebuild."))
+    return (repair_xref(pdf, password_cb, password_data));
+  else
+    return (false);
 }
 
 
@@ -2208,7 +2306,7 @@ repair_xref(
     pdfio_password_cb_t password_cb,	// I - Password callback or `NULL` for none
     void                *password_data)	// I - Password callback data, if any
 {
-  char		line[16384],		// Line from file
+  char		line[1024],		// Line from file
 		*ptr;			// Pointer into line
   off_t		line_offset;		// Offset in file
   intmax_t	number;			// Object number
@@ -2216,16 +2314,22 @@ repair_xref(
   size_t	i;			// Looping var
   size_t	num_sobjs = 0;		// Number of object streams
   pdfio_obj_t	*sobjs[16384];		// Object streams to load
+  pdfio_dict_t	*backup_trailer = NULL;	// Backup trailer dictionary
+  pdfio_obj_t	*pages_obj;		// Pages object
 
 
-  // Let caller know something is wrong...
-  _pdfioFileError(pdf, "WARNING: Cross-reference table is damaged, attempting to rebuild.");
+  // Clear trailer data...
+  pdf->trailer_dict = NULL;
+  pdf->root_obj     = NULL;
+  pdf->info_obj     = NULL;
+  pdf->pages_obj    = NULL;
+  pdf->encrypt_obj  = NULL;
 
-  // Read from the beginning of the file, looking for
+  // Read from the beginning of the file, looking for objects...
   if ((line_offset = _pdfioFileSeek(pdf, 0, SEEK_SET)) < 0)
     return (false);
 
-  while (_pdfioFileGets(pdf, line, sizeof(line)))
+  while (_pdfioFileGets(pdf, line, sizeof(line), true))
   {
     // See if this is the start of an object...
     if (line[0] >= '1' && line[0] <= '9')
@@ -2242,43 +2346,75 @@ repair_xref(
 	  pdfio_obj_t	*obj;		// Object
 	  _pdfio_token_t tb;		// Token buffer/stack
 
-          PDFIO_DEBUG("OBJECT %ld %d at offset %ld\n", (long)number, generation, (long)line_offset);
+          PDFIO_DEBUG("repair_xref: OBJECT %ld %d at offset %ld\n", (long)number, generation, (long)line_offset);
 
-	  if ((obj = add_obj(pdf, (size_t)number, (unsigned short)generation, line_offset)) == NULL)
+          if ((obj = pdfioFileFindObj(pdf, (size_t)number)) != NULL)
+          {
+            obj->offset = line_offset;
+          }
+          else if ((obj = add_obj(pdf, (size_t)number, (unsigned short)generation, line_offset)) == NULL)
 	  {
 	    _pdfioFileError(pdf, "Unable to allocate memory for object.");
 	    return (false);
 	  }
 
+          if (ptr[3])
+          {
+	    // Probably the start of the object dictionary, rewind the file so
+	    // we can read it...
+	    _pdfioFileSeek(pdf, line_offset + (ptr - line + 3), SEEK_SET);
+	  }
+
 	  _pdfioTokenInit(&tb, pdf, (_pdfio_tconsume_cb_t)_pdfioFileConsume, (_pdfio_tpeek_cb_t)_pdfioFilePeek, pdf);
 
 	  if (!_pdfioValueRead(pdf, obj, &tb, &obj->value, 0))
 	  {
-	    _pdfioFileError(pdf, "Unable to read cross-reference stream dictionary.");
-	    return (false);
+	    if (!_pdfioFileError(pdf, "WARNING: Unable to read object dictionary/value."))
+	      return (false);
+	    else
+	      continue;
 	  }
 
-	  if (_pdfioTokenGet(&tb, line, sizeof(line)) && strcmp(line, "stream"))
+	  if (_pdfioTokenGet(&tb, line, sizeof(line)))
 	  {
 	    const char *type = pdfioObjGetType(obj);
 					// Object type
 
 	    _pdfioTokenFlush(&tb);
-	    obj->stream_offset = _pdfioFileTell(pdf);
 
-	    if (type && !strcmp(type, "ObjStm") && num_sobjs < (sizeof(sobjs) / sizeof(sobjs[0])))
-	    {
-	      sobjs[num_sobjs] = obj;
-	      num_sobjs ++;
-	    }
-
-            if (type && !strcmp(type, "XRef") && !pdf->trailer_dict)
+            if (type && !strcmp(line, "stream"))
             {
-              // Save the trailer dictionary...
-	      pdf->trailer_dict = pdfioObjGetDict(obj);
-	      pdf->encrypt_obj  = pdfioDictGetObj(pdf->trailer_dict, "Encrypt");
-	      pdf->id_array     = pdfioDictGetArray(pdf->trailer_dict, "ID");
-            }
+              // Possible object or XRef stream...
+	      obj->stream_offset = _pdfioFileTell(pdf);
+
+	      if (!strcmp(type, "ObjStm") && num_sobjs < (sizeof(sobjs) / sizeof(sobjs[0])))
+	      {
+	        PDFIO_DEBUG("repair_xref: Object stream...\n");
+		sobjs[num_sobjs] = obj;
+		num_sobjs ++;
+	      }
+
+	      if (!strcmp(type, "XRef") && !pdf->trailer_dict)
+	      {
+		// Save the trailer dictionary...
+	        PDFIO_DEBUG("repair_xref: XRef stream...\n");
+		pdf->trailer_dict = pdfioObjGetDict(obj);
+		pdf->encrypt_obj  = pdfioDictGetObj(pdf->trailer_dict, "Encrypt");
+		pdf->id_array     = pdfioDictGetArray(pdf->trailer_dict, "ID");
+	      }
+	    }
+	    else if (type && !strcmp(line, "endobj"))
+	    {
+	      // Possible catalog or pages object...
+	      if (!strcmp(type, "Catalog"))
+	      {
+	        PDFIO_DEBUG("repair_xref: Catalog (root) object...\n");
+	        if (!backup_trailer)
+	          backup_trailer = pdfioDictCreate(pdf);
+
+	        pdfioDictSetObj(backup_trailer, "Root", obj);
+	      }
+	    }
 	  }
 	}
       }
@@ -2289,6 +2425,8 @@ repair_xref(
       _pdfio_token_t tb;		// Token buffer/stack
       _pdfio_value_t trailer;		// Trailer
 
+      PDFIO_DEBUG("repair_xref: line=\"%s\"\n", line);
+
       if (line[7])
       {
 	// Probably the start of the trailer dictionary, rewind the file so
@@ -2296,7 +2434,7 @@ repair_xref(
 	_pdfioFileSeek(pdf, line_offset + 7, SEEK_SET);
       }
 
-      PDFIO_DEBUG("TRAILER at offset %ld\n", (long)line_offset);
+      PDFIO_DEBUG("repair_xref: TRAILER at offset %ld\n", (long)line_offset);
 
       _pdfioTokenInit(&tb, pdf, (_pdfio_tconsume_cb_t)_pdfioFileConsume, (_pdfio_tpeek_cb_t)_pdfioFilePeek, pdf);
       if (!_pdfioValueRead(pdf, NULL, &tb, &trailer, 0))
@@ -2312,10 +2450,12 @@ repair_xref(
 
       _pdfioTokenFlush(&tb);
 
-      if (!pdf->trailer_dict)
+      if (_pdfioDictGetValue(trailer.value.dict, "Root"))
       {
 	// Save the trailer dictionary and grab the root (catalog) and info
 	// objects...
+	PDFIO_DEBUG("repair_xref: Using this trailer dictionary.\n");
+
 	pdf->trailer_dict = trailer.value.dict;
 	pdf->encrypt_obj  = pdfioDictGetObj(pdf->trailer_dict, "Encrypt");
 	pdf->id_array     = pdfioDictGetArray(pdf->trailer_dict, "ID");
@@ -2326,11 +2466,18 @@ repair_xref(
     line_offset = _pdfioFileTell(pdf);
   }
 
+  PDFIO_DEBUG("repair_xref: Stopped at line_offset=%lu\n", (unsigned long)line_offset);
+
+  if (!pdf->trailer_dict && backup_trailer)
+    pdf->trailer_dict = backup_trailer;
+
   // If the trailer contains an Encrypt key, try unlocking the file...
   if (pdf->encrypt_obj && !_pdfioCryptoUnlock(pdf, password_cb, password_data))
     return (false);
 
   // Load any stream objects...
+  PDFIO_DEBUG("repair_xref: Found %lu stream objects.\n", (unsigned long)num_sobjs);
+
   for (i = 0; i < num_sobjs; i ++)
   {
     if (!load_obj_stream(sobjs[i]))
@@ -2349,8 +2496,16 @@ repair_xref(
 
   PDFIO_DEBUG("repair_xref: Root=%p(%lu)\n", pdf->root_obj, (unsigned long)pdf->root_obj->number);
 
+  if ((pages_obj = pdfioDictGetObj(pdfioObjGetDict(pdf->root_obj), "Pages")) == NULL)
+  {
+    _pdfioFileError(pdf, "Missing Pages object.");
+    return (false);
+  }
+
+  PDFIO_DEBUG("repair_xref: Pages=%p(%lu)\n", pages_obj, (unsigned long)pages_obj->number);
+
   // Load pages...
-  return (load_pages(pdf, pdfioDictGetObj(pdfioObjGetDict(pdf->root_obj), "Pages"), 0));
+  return (load_pages(pdf, pages_obj, 0));
 }
 
 
@@ -2500,12 +2655,16 @@ write_trailer(pdfio_file_t *pdf)	// I - PDF file
             buffer[2] = (obj->offset >> 8) & 255;
             buffer[3] = obj->offset & 255;
             break;
+#ifdef _WIN32
+	default :
+#endif // _WIN32
         case 4 :
             buffer[1] = (obj->offset >> 24) & 255;
             buffer[2] = (obj->offset >> 16) & 255;
             buffer[3] = (obj->offset >> 8) & 255;
             buffer[4] = obj->offset & 255;
             break;
+#ifndef _WIN32 // Windows off_t is 32-bits?!?
         case 5 :
             buffer[1] = (obj->offset >> 32) & 255;
             buffer[2] = (obj->offset >> 24) & 255;
@@ -2540,6 +2699,7 @@ write_trailer(pdfio_file_t *pdf)	// I - PDF file
             buffer[7] = (obj->offset >> 8) & 255;
             buffer[8] = obj->offset & 255;
             break;
+#endif // !_WIN32
       }
 
       if (!pdfioStreamWrite(xref_st, buffer, offsize + 2))
