@@ -172,7 +172,7 @@ _pdfioValueDecrypt(pdfio_file_t   *pdf,	// I - PDF file
 	// Copy the decrypted string back to the value and adjust the length...
 	memcpy(v->value.binary.data, temp, templen);
 
-	if (pdf->encryption >= PDFIO_ENCRYPTION_AES_128)
+	if (pdf->encryption >= PDFIO_ENCRYPTION_AES_128 && temp[templen - 1] <= templen)
 	  v->value.binary.datalen = templen - temp[templen - 1];
 	else
 	  v->value.binary.datalen = templen;
@@ -183,20 +183,48 @@ _pdfioValueDecrypt(pdfio_file_t   *pdf,	// I - PDF file
     case PDFIO_VALTYPE_STRING :
         // Decrypt regular string...
         templen = strlen(v->value.string);
-	if (templen > (sizeof(temp) - 33))
+	if (templen > (PDFIO_MAX_STRING - 1))
 	{
 	  _pdfioFileError(pdf, "Unable to read encrypted string - too long.");
 	  return (false);
 	}
+	else if ((temp = (uint8_t *)_pdfioStringAllocBuffer(pdf)) == NULL)
+	{
+	  _pdfioFileError(pdf, "Unable to read encrypted binary string - out of memory.");
+	  return (false);
+	}
 
 	ivlen = templen;
 	if ((cb = _pdfioCryptoMakeReader(pdf, obj, &ctx, (uint8_t *)v->value.string, &ivlen)) == NULL)
 	  return (false);
 
 	templen = (cb)(&ctx, temp, (uint8_t *)v->value.string + ivlen, templen - ivlen);
+
+	if (pdf->encryption >= PDFIO_ENCRYPTION_AES_128 && temp[templen - 1] <= templen)
+	  templen -= temp[templen - 1];
+
 	temp[templen] = '\0';
 
-        if ((timeval = get_date_time((char *)temp)) != 0)
+        if ((templen & 1) == 0 && (!memcmp(temp, "\376\377", 2) || !memcmp(temp, "\377\376", 2)))
+        {
+          // Convert UTF-16 to UTF-8...
+          char	utf8[4096];		// Temporary string
+
+          _pdfio_utf16cpy(utf8, temp, templen, sizeof(utf8));
+
+	  if ((timeval = get_date_time((char *)utf8)) != 0)
+	  {
+	    // Change the type to date...
+	    v->type       = PDFIO_VALTYPE_DATE;
+	    v->value.date = timeval;
+	  }
+	  else
+	  {
+	    // Copy the decrypted string back to the value...
+	    v->value.string = pdfioStringCreate(pdf, utf8);
+	  }
+        }
+        else if ((timeval = get_date_time((char *)temp)) != 0)
         {
           // Change the type to date...
           v->type       = PDFIO_VALTYPE_DATE;
@@ -207,6 +235,8 @@ _pdfioValueDecrypt(pdfio_file_t   *pdf,	// I - PDF file
           // Copy the decrypted string back to the value...
 	  v->value.string = pdfioStringCreate(pdf, (char *)temp);
 	}
+
+        _pdfioStringFreeBuffer(pdf, (char *)temp);
         break;
   }
 
@@ -367,18 +397,22 @@ _pdfioValueRead(pdfio_file_t   *pdf,	// I - PDF file
 
     ret = v;
   }
-  else if ((timeval = get_date_time(token + 1)) != 0)
-  {
-    v->type       = PDFIO_VALTYPE_DATE;
-    v->value.date = timeval;
-    ret           = v;
-  }
   else if (token[0] == '(')
   {
-    // String
-    v->type         = PDFIO_VALTYPE_STRING;
-    v->value.string = pdfioStringCreate(pdf, token + 1);
-    ret           = v;
+    if ((timeval = get_date_time(token + 1)) != 0)
+    {
+      // Date
+      v->type       = PDFIO_VALTYPE_DATE;
+      v->value.date = timeval;
+      ret           = v;
+    }
+    else
+    {
+      // String
+      v->type         = PDFIO_VALTYPE_STRING;
+      v->value.string = pdfioStringCreate(pdf, token + 1);
+      ret           = v;
+    }
   }
   else if (token[0] == '/')
   {
